package org.processmining.datapetrinets.datadiscovery;

import java.io.File;
import java.io.IOException;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;

import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.models.FunctionEstimator.DecisionTreeBasedFunctionEstimator;
import org.processmining.models.FunctionEstimator.FunctionEstimation;
import org.processmining.models.FunctionEstimator.Type;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.petrinetwithdata.decisionminer.DecisionMining;

import com.google.common.base.Joiner;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;

public abstract class AbstractDecisionRuleDiscovery implements DecisionRuleDiscovery {

	public interface ProjectedEvent {

		Transition getTransition();

		Object getAttributeValue(String attributeName);

		Set<String> getAttributes();

	}

	public interface ProjectedTrace extends Iterable<ProjectedEvent> {
	}

	public interface ProjectedLog extends Iterable<ProjectedTrace> {
		Set<String> getAttributes();

		Object getInitialValue(String attributeName);
	}

	public static final class DecisionRuleResultDecisionTreeImpl implements DecisionRule {

		private final Map<Object, FunctionEstimation> estimation;
		private final Place place;
		private final String overallInfo;

		public DecisionRuleResultDecisionTreeImpl(Place place, Map<Object, FunctionEstimation> estimation,
				DecisionTreeBasedFunctionEstimator estimator) {
			this.place = place;
			this.estimation = estimation;
			this.overallInfo = estimator.toString();
		}

		public Map<Transition, GuardExpression> getRules() {
			HashMap<Transition, GuardExpression> rules = Maps.newHashMapWithExpectedSize(estimation.size());
			for (Entry<Object, FunctionEstimation> entry : estimation.entrySet()) {
				rules.put((Transition) entry.getKey(), entry.getValue().getExpression());
			}
			return rules;
		}

		public Map<Transition, Double> getDecisionTreeConfidence() {
			HashMap<Transition, Double> confidence = Maps.newHashMapWithExpectedSize(estimation.size());
			for (Entry<Object, FunctionEstimation> entry : estimation.entrySet()) {
				confidence.put((Transition) entry.getKey(), entry.getValue().getQualityMeasure());
			}
			return confidence;
		}

		public String toString() {
			return estimation.toString();
		}

		public String getOverallInfo() {
			return overallInfo;
		}

		public Place getPlace() {
			return place;
		}

	}

	protected static final class DecisionRulesResultImpl implements DecisionRules {

		private final Map<Place, DecisionRule> ruleMapping;

		private DecisionRulesResultImpl(Map<Place, DecisionRule> ruleMapping) {
			this.ruleMapping = ruleMapping;
		}

		public Map<Place, DecisionRule> getRules() {
			return ruleMapping;
		}

		public String toString() {
			return Joiner.on(",\n").withKeyValueSeparator("=").join(ruleMapping);
		}

	}

	protected final ProjectedLog projectedLog;
	protected final Map<String, Type> attributeType;
	protected final Map<String, Set<String>> literalValues;
	protected final DecisionTreeConfig config;
	private final int numInstancesEstimate;

	public AbstractDecisionRuleDiscovery(DecisionTreeConfig config, Iterable<ProjectedTrace> projectedLog,
			int numInstancesEstimate) {
		this(config, projectedLog, extractAttributeInformation(projectedLog, config),
				getLiteralValuesMap(projectedLog, config), numInstancesEstimate); //TODO join both literal values map and attribute type info in one data structure
	}

	private static Map<String, Set<String>> getLiteralValuesMap(Iterable<ProjectedTrace> projectedLog,
			DecisionTreeConfig config) {
		Map<String, Set<String>> retValue = new HashMap<>();
		for (ProjectedTrace trace : projectedLog) {
			for (ProjectedEvent event : trace) {
				for (String attributeKey : event.getAttributes()) {
					Object value = event.getAttributeValue(attributeKey);
					if (value instanceof String) {
						String varName = escapeAttributeName(attributeKey);
						Set<String> literalValues = retValue.get(varName);
						if (literalValues == null) {
							literalValues = new TreeSet<>();
							retValue.put(varName, literalValues);
						}
						literalValues.add((String) value);
					}
				}
			}
		}
		return retValue;
	}

	private static Map<String, Type> extractAttributeInformation(Iterable<ProjectedTrace> projectedLog,
			DecisionTreeConfig config) {
		HashMap<String, Type> retValue = new HashMap<>();
		for (ProjectedTrace trace : projectedLog) {
			for (ProjectedEvent event : trace) {
				for (String attributeKey : event.getAttributes()) {
					String fixedVarName = escapeAttributeName(attributeKey);
					// Uses the first occurrence and assumes data types to be consistent!
					if (!retValue.containsKey(fixedVarName)) {
						Object value = event.getAttributeValue(attributeKey);
						Type classType = generateDataElement(value);
						if (classType != null) {
							retValue.put(fixedVarName, classType);
						}
					}
				}
			}

		}
		return retValue;
	}

	private static Type generateDataElement(Object value) {
		if (value instanceof Boolean) {
			return Type.BOOLEAN;
		} else if (value instanceof Double || value instanceof Float) {
			return Type.CONTINUOS;
		} else if (value instanceof Long || value instanceof Integer) {
			return Type.DISCRETE;
		} else if (value instanceof Date) {
			return Type.TIMESTAMP;
		} else if (value instanceof String) {
			return Type.LITERAL;
		} else {
			// Non "primitive" types like list are not supported
			return null;
		}
	}

	public AbstractDecisionRuleDiscovery(DecisionTreeConfig config, final Iterable<ProjectedTrace> projectedLog,
			Map<String, Type> attributeType, Map<String, Set<String>> literalValues, int numInstancesEstimate) {
		this.config = config;
		if (projectedLog instanceof ProjectedLog) {
			this.projectedLog = (ProjectedLog) projectedLog; //TODO change after all callers have been updated to avoid Hudson issues	
		} else {
			this.projectedLog = new ProjectedLog() {

				public Iterator<ProjectedTrace> iterator() {
					return projectedLog.iterator();
				}

				public Object getInitialValue(String attributeName) {
					return null;
				}

				public Set<String> getAttributes() {
					return ImmutableSet.of();
				}
			};
		}
		this.attributeType = attributeType;
		this.literalValues = literalValues;
		this.numInstancesEstimate = numInstancesEstimate;
	}

	public final DecisionRules discoverRules(Petrinet net) throws RuleDiscoveryException {
		final Map<Place, DecisionRule> ruleMapping = new HashMap<>();
		for (Place place : net.getPlaces()) {
			if (net.getOutEdges(place).size() > 1) {
				ruleMapping.put(place, discoverRulesForPlace(net, place));
			}
		}
		return new DecisionRulesResultImpl(ruleMapping);
	}

	protected int getNumInstanceEstimate() {
		return numInstancesEstimate;
	}

	protected static Set<Transition> getTransitionPostset(PetrinetGraph net, Place place) {
		Set<Transition> postSet = Sets.newHashSetWithExpectedSize(net.getOutEdges(place).size());
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> arc : net.getOutEdges(place)) {
			postSet.add((Transition) arc.getTarget());
		}
		return postSet;
	}

	protected final DecisionRule doDiscoverDecisionTree(Place place, Set<Transition> transitions,
			DecisionTreeBasedFunctionEstimator estimator) throws RuleDiscoveryException {

		estimator.setBinarySplit(config.isBinarySplit());
		estimator.setCrossValidate(config.isCrossValidate());
		estimator.setUnpruned(config.isUnpruned());

		try {
			final Map<String, Object> initialAttributeValues = getInitialValues(projectedLog);
			final Map<String, Object> currentAttributeValues = new HashMap<>();
			for (ProjectedTrace trace : projectedLog) {
				currentAttributeValues.putAll(initialAttributeValues);
				for (ProjectedEvent e : trace) {
					if (config.isMinePrimeGuards()) {
						for (String attributeKey : e.getAttributes()) {
							// NULL is used as marker for missing values
							currentAttributeValues.put(escapeAttributeName(attributeKey).concat("'"),
									e.getAttributeValue(attributeKey));
						}
					}
					final Transition transition = e.getTransition();
					if (transitions.contains(transition)) {
						estimator.addInstance(currentAttributeValues, transition, 1.0f);
					}
					// Update current values					
					for (String attributeKey : e.getAttributes()) {
						// NULL is used as marker for missing values
						currentAttributeValues.put(escapeAttributeName(attributeKey),
								e.getAttributeValue(attributeKey));
					}
				}
				currentAttributeValues.clear();
			}
		} catch (Exception ex) {
			throw new RuleDiscoveryException(ex);
		}

		if (config.isSaveInstances()) {
			try {
				estimator.saveInstances(new File(place.getLabel() + ".arff"));
			} catch (IOException e) {
				throw new RuleDiscoveryException("Cannot save instances", e);
			}
		}

		// Each place gets its own minNumObj configuration
		int minNumObj = Math.max(2,
				(int) Math.floor((config.getMinPercentageObjectsOnLeaf() * estimator.getNumInstances())));
		estimator.setMinNumObj(minNumObj);

		try {
			Map<Object, FunctionEstimation> estimation = estimator.getFunctionEstimation(null);
			return new DecisionRuleResultDecisionTreeImpl(place, estimation, estimator);
		} catch (Exception ex) {
			throw new RuleDiscoveryException("An error occured while discovering a rule for place " + place, ex);
		} catch (StackOverflowError error) {
			throw new RuleDiscoveryException(
					"One of the rules discovered at place '" + place
							+ "' has too many levels of nesting! A stack overflow occured while processing this rule. Probably the used classifier returned a tree with too many leafs.",
					error);
		}
	}

	private HashMap<String, Object> getInitialValues(ProjectedLog projectedLog) {
		HashMap<String, Object> values = new HashMap<>();
		for (String attribute : projectedLog.getAttributes()) {
			values.put(escapeAttributeName(attribute), projectedLog.getInitialValue(attribute));
		}
		return values;
	}

	public static String escapeAttributeName(String attribute) {
		//TODO find something better than this, unfortunately WEKA is rather strict on attribute names
		return DecisionMining.fixVarName(attribute);
	}

}